local BaikCore = LibStub("AceAddon-3.0"):GetAddon("BaikCore")
local Assert = BaikCore.Assert
local Database = BaikCore.Database
local Option = BaikCore.Option
local OrderNumber = BaikCore.OrderNumber
local Ui = BaikCore.Ui

-- Add Module
local VertexColor = BaikCore:NewModule("VertexColor")
BaikCore.VertexColor = VertexColor

-- Constants
local SUFFIX_DB_ENABLED = "%s_color_on"
local SUFFIX_DB_COLOR = "%s_color"
local SUFFIX_DB_CLASS = "%s_color_class"
local SUFFIX_DB_CLASS_ALPHA = "%s_color_class_alpha"

local PREFIX_DB_GLOBAL = "global"
local GLOBAL_DB_ENABLED = string.format(SUFFIX_DB_ENABLED, PREFIX_DB_GLOBAL)
local GLOBAL_DB_COLOR = string.format(SUFFIX_DB_COLOR, PREFIX_DB_GLOBAL)
local GLOBAL_DB_CLASS = string.format(SUFFIX_DB_CLASS, PREFIX_DB_GLOBAL)
local GLOBAL_DB_CLASS_ALPHA = string.format(SUFFIX_DB_CLASS_ALPHA, PREFIX_DB_GLOBAL)

VertexColor.DEFAULT_ENABLE = false
VertexColor.DEFAULT_COLOR = Ui:CreateColor(1, 1, 1, 1)
VertexColor.DEFAULT_CLASS = false
VertexColor.DEFAULT_CLASS_ALPHA = 1

local CLASS_COLOR = nil

-- Private Variables
local colorFunc = {}

-- Private Methods
local function GetDbEnable(dbPrefix)
    return string.format(SUFFIX_DB_ENABLED, dbPrefix)
end

local function GetDbColor(dbPrefix)
    return string.format(SUFFIX_DB_COLOR, dbPrefix)
end

local function GetDbClass(dbPrefix)
    return string.format(SUFFIX_DB_CLASS, dbPrefix)
end

local function GetDbAlpha(dbPrefix)
    return string.format(SUFFIX_DB_CLASS_ALPHA, PREFIX_DB_GLOBAL)
end

local function GetEnabled(dbPrefix)
    local globalEnabled = Database:Get(GetDbEnable(PREFIX_DB_GLOBAL))

    return globalEnabled or Database:Get(GetDbEnable(dbPrefix))
end

local function GetColor(dbPrefix)
    local globalEnabled = Database:Get(GetDbEnable(PREFIX_DB_GLOBAL))
    local localPrefix = globalEnabled and PREFIX_DB_GLOBAL or dbPrefix

    if Database:Get(GetDbClass(localPrefix)) then
        local classColor = Ui:GetClassColor("player")
        classColor.a = Database:Get(GetDbAlpha(localPrefix))

        return classColor
    end

    return Database:Get(GetDbColor(localPrefix))
end

local function ColorFrameInner(frame, dbPrefix)
    local color = GetColor(dbPrefix)
    frame:SetVertexColor(color.r, color.g, color.b, color.a)
end

local function SyncAll()
    for _, func in pairs(colorFunc) do
        if func ~= SyncAll then
            func()
        end
    end
end

local function GetEnableOpts(dbPrefix, func, orderNumber)
    colorFunc[dbPrefix] = func
    local dbEnable = GetDbEnable(dbPrefix)

    return {
        name = "Enable",
        type = "toggle",
        order = orderNumber:GetNext(),
        disabled = function(...)
            local globalDb = GetDbEnable(PREFIX_DB_GLOBAL)

            return dbPrefix ~= PREFIX_DB_GLOBAL and Database:Get(globalDb)
        end,
        get = function(...)
            return Database:Get(dbEnable)
        end,
        set = function(self, value, ...)
            Database:Set(dbEnable, value)

            colorFunc[dbPrefix]()
        end
    }
end

local function GetColorOpts(dbPrefix, orderNumber)
    local dbColor = GetDbColor(dbPrefix)

    return {
        name = "Color",
        type = "color",
        order = orderNumber:GetNext(),
        hasAlpha = true,
        disabled = function(...)
            local globalDb = GetDbEnable(PREFIX_DB_GLOBAL)

            -- Apply different logic for global
            if dbPrefix == PREFIX_DB_GLOBAL then
                local globalDbClass = GetDbClass(PREFIX_DB_GLOBAL)
                return not Database:Get(globalDb) or Database:Get(globalDbClass)
            end

            -- Check if global or local class is enabled
            local localDbClass = GetDbClass(PREFIX_DB_GLOBAL)
            return Database:Get(globalDb) or Database:Get(localDbClass)
        end,
        get = function(...)
            local color = Database:Get(dbColor)
            return color.r, color.g, color.b, color.a
        end,
        set = function(self, r, g, b, a, ...)
            Assert:Number(r)
            Assert:Number(g)
            Assert:Number(b)
            Assert:Number(a)
            Database:Set(dbColor, Ui:CreateColor(r, g, b, a))

            colorFunc[dbPrefix]()
        end
    }
end

local function GetClassOpts(dbPrefix, orderNumber)
    local dbClass = GetDbClass(dbPrefix)

    return {
        name = "Class Color",
        type = "toggle",
        order = orderNumber:GetNext(),
        disabled = function(...)
            local globalDb = GetDbEnable(PREFIX_DB_GLOBAL)

            -- Apply different logic for global
            if dbPrefix == PREFIX_DB_GLOBAL then
                return not Database:Get(globalDb)
            end

            -- Check if global
            local localDb = GetDbEnable(dbPrefix)
            return Database:Get(globalDb) or not Database:Get(localDb)
        end,
        get = function(...)
            return Database:Get(dbClass)
        end,
        set = function(self, value, ...)
            Assert:Boolean(value)
            Database:Set(dbClass, value)

            colorFunc[dbPrefix]()
        end
    }
end

local function GetAlphaOpts(dbPrefix, orderNumber)
    local dbAlpha = GetDbAlpha(dbPrefix)
    return {
        name = "Class Alpha",
        type = "range",
        order = orderNumber:GetNext(),
        min = 0,
        max = 1,
        isPercent = true,
        disabled = function(...)
            local globalDb = GetDbEnable(PREFIX_DB_GLOBAL)

            -- Apply different logic for global
            if dbPrefix == PREFIX_DB_GLOBAL then
                local globalDbClass = GetDbClass(PREFIX_DB_GLOBAL)
                return not (Database:Get(globalDb) and Database:Get(globalDbClass))
            end

            -- Check if global or local class is enabled
            local localDbClass = GetDbClass(dbPrefix)
            return Database:Get(globalDb) or not Database:Get(localDbClass)
        end,
        get = function(...)
            return Database:Get(dbAlpha)
        end,
        set = function(self, value, ...)
            Assert:Number(value)
            Database:Set(dbAlpha, value)
            colorFunc[dbPrefix]()
        end
    }
end

-- Public API
function VertexColor:ColorFrame(frame, dbPrefix)
    Assert:Table(frames)
    Assert:String(dbPrefix)

    if not GetEnabled(dbPrefix) then
        return
    end

    ColorFrameInner(frame, dbPrefix)
end

function VertexColor:ColorFrameAll(frames, dbPrefix)
    Assert:Table(frames)
    Assert:String(dbPrefix)

    if not GetEnabled(dbPrefix) then
        return
    end

    for _, frame in pairs(frames) do
        ColorFrameInner(frame, dbPrefix)
    end
end

function VertexColor:Sync(dbPrefix)
    Assert:String(dbPrefix)
    -- Ensure we have frames to sync
    if not colorFunc[dbPrefix] then
        Log:w(string.format("%s does not have a color sync function", dbValue))
        return
    end

    colorFunc[dbValue]()
end

function VertexColor:InsertOpts(argTbl, dbPrefix, func, orderNumber)
    Assert:Table(argTbl)
    Assert:String(dbPrefix)
    Assert:Function(func)

    if not orderNumber then
        orderNumber = OrderNumber:Create()
    end

    argTbl["enabled"] = GetEnableOpts(dbPrefix, func, orderNumber)
    argTbl["color"] = GetColorOpts(dbPrefix, orderNumber)
    argTbl["class"] = GetClassOpts(dbPrefix, orderNumber)
    argTbl["class_alpha"] = GetAlphaOpts(dbPrefix, orderNumber)
end

function VertexColor:GetOpts(dbPrefix, func)
    local opts = {
        name = "Art",
        type = "group",
        args = {
        }
    }
    self:InsertOpts(opts.args, dbPrefix, func)

    return opts
end


function VertexColor:InitDb(dbPrefix)
    Assert:String(dbPrefix)

    Database:InitIfEmpty(GetDbEnable(dbPrefix), self.DEFAULT_ENABLE)
    Database:InitIfEmpty(GetDbColor(dbPrefix), self.DEFAULT_COLOR)
    Database:InitIfEmpty(GetDbClass(dbPrefix), self.DEFAULT_CLASS)
    Database:InitIfEmpty(GetDbAlpha(dbPrefix), self.DEFAULT_CLASS_ALPHA)
end

-- Ace3 Callbacks
function VertexColor:OnInitialize()
    self:InitDb(PREFIX_DB_GLOBAL)
    Option:AddOption(Option.NAME, "vertex_color", self:GetOpts(PREFIX_DB_GLOBAL, SyncAll))
end

function VertexColor:OnEnable()
end

function VertexColor:OnDisable()
end
